
Require Bool.
Require Arith.
Require Min.

Load Document.

(* Constructor for the (hypothetical) output of the "meaning function",
   which would be a string, annotated with some properties. *)
Definition pair_output := (pair properties string).

(* Utility function for increasing the "U" level, and proving
   that we don't increment it past 3. *)
Definition inc_U :=
  [x:nat] (min (S x) (S (S (S O))) ).
Goal (a:nat) (le (inc_U a) (S (S (S O))) ).
  Intros.
  Unfold inc_U.
  Exact (le_min_r (S a) (S (S (S O))) ).
Save inc_U_le_three.
Goal (le O (S (S (S O))) ).
Auto.
Save O_le_three.

(* the "meaning" of a document. *)
(* XXX note that one-field (purely functional) update of records,
   as seen in various flavors of ocaml and Haskell, would make
   this much nicer... *)
Fixpoint meaning [d:document] : properties -> (list (properties * string)) :=
 [c:properties]
  Cases d of
    (d_list doclist) => (nil (properties * string))
  | (d_blank_chars s) => (cons (pair_output c s)
                               (nil (properties * string)))
  | (d_nonblank_chars s) => (cons (pair_output c s)
                               (nil (properties * string)))
  | (d_B d') => (meaning d'
     (mk_properties true (em c) (i c) (s c) (tt c)
      (u c) (u_no_more_than_three c) (size c) (color c)))
  | (d_EM d') => (meaning d'
     (mk_properties (b c) (em c)
       (* "if s is not set, then invert i, else do nothing." *)
       (ifb (s c) (i c) (negb (i c)))
       (s c) (tt c)
      (u c) (u_no_more_than_three c) (size c) (color c)))
  | (d_I d') => (meaning d'
     (mk_properties (b c) (em c) true (s c) (tt c)
      (u c) (u_no_more_than_three c) (size c) (color c)))
  | (d_PL d') => (meaning d'
     (mk_properties false false false false false
      O O_le_three (size c) (color c)))
  | (d_S d') => (meaning d'
     (mk_properties (b c) false (i c) true (tt c)
      (u c) (u_no_more_than_three c) (size c) (color c)))
  | (d_TT d') => (meaning d'
     (mk_properties (b c) (em c) (i c) (s c) true
      (u c) (u_no_more_than_three c) (size c) (color c)))
  | (d_U d') => (meaning d'
     (mk_properties (b c) (em c) (i c) (s c) (tt c)
      (inc_U (u c))
      (inc_U_le_three (u c))
      (size c) (color c)))
  | (d_size sz d') => (meaning d'
     (mk_properties (b c) (em c) (i c) (s c) (tt c)
      (u c) (u_no_more_than_three c) sz (color c)))
  | (d_color clr d') => (meaning d'
     (mk_properties (b c) (em c) (i c) (s c) (tt c)
      (u c) (u_no_more_than_three c) (size c) clr))
  end
  with meaning_l [dl:document_list] : properties ->
                     (list (properties * string)) :=
  [c:properties]
  Cases dl of
    (doc_list d1) => (flat_map ([foo:document] (meaning foo c)) d1)
  end.

(* root contexts have most flags set to "false." *)
Definition root_context :=
  [c:properties]
    ((b c) = false) /\
    ((em c) = false) /\
    ((i c) = false) /\
    ((s c) = false) /\
    ((tt c) = false) /\
    ((u c) = O).

(* the type of transformations which don't change the "meaning"
   of the document. *)
Inductive valid_transform : document -> document -> Prop :=
  same_meaning :
    (a,b:document)
      ((c:properties) (root_context c) ->
        ((meaning a c) = (meaning b c))) ->
      (valid_transform a b).



